/* This file is in the public domain.
   Default styling rules for GNU poke.  Bright background version.  */

/* .logo .copyright .type .struct-type-name .array */

.logo { text-decoration: blink; }
.error { font-weight: bold; color : red; }
.error-filename { text-decoration: underline; }
.error-location { text-decoration: underline; }
.warning { color : yellow; }
.prompt { font-weight : bold; }
.offset { color : yellow; }
.special { color : white; }
.string { color : brown; }
.any { color : yellow; }
.struct-field-name { text-decoration : underline; }
.integer { color : green; }

.table-header { text-decoration: underline; }

.setting-integer { color : green; }
.setting-string { color : red; }
.setting-boolean { color : yellow; }

.dump-ruler { text-decoration: underline; }
.dump-address { color : green; }
.dump-ascii { color : brown; }
.dump-unknown { color : yellow; }

.diff-thunk-header { font-weight: bold; }
.diff-minus { color : red; }
.diff-plus { color : green; }

/* Classes for pickles in pickles/  */

.leb128 { color : grey; }

/* Classes for ISA-related stuff.  */
.insn-mnemonic { color : yellow; }
.insn-register { color : yellow; }

/* Classes for the pvm disassembly.  */

.pvm-comment { color: brown; font-style: italic; }
.pvm-punctuation { color: white; }
.pvm-instruction { color: cyan; }
.pvm-label { color: yellow;   }
.pvm-register { color: green; }
.pvm-number { color: lightgreen; }

.pager-more { color: white; background-color: black; }

html {
    font-family: Menlo, Monaco, "Courier New", monospace;
}
